open Syntax 
open Util


(* Collect all defining occurences of nonces and variables of a process *)
let rec collectnv (p:process) : ((nonce * label) list * var list) = match p with 
  | PZero _  -> ([],[])
  | PNew (_, n, l, p') -> let (ns,vs) = collectnv p' in ((n,l)::ns,vs)
  | POut (_,_,p') -> collectnv p' 
  | PIn (_,x, p') -> let (ns,vs) = collectnv p' in (ns,x::vs)
  | PPar (_,p1,p2) -> 
      let (ns,vs) = collectnv p1 in 
      let (ns',vs') = collectnv p2 in 
      (ns @ ns', vs @ vs')
  | PRepl (_,p') -> collectnv p' 
  | PLet (_,x,_,p1,p2) -> 
      let (ns,vs) = collectnv p1 in 
      let (ns',vs') = collectnv p2 in 
      (ns @ ns', x :: vs @ vs')
  | PIf (_,_,_,p1,p2) -> 
      let (ns,vs) = collectnv p1 in 
      let (ns',vs') = collectnv p2 in 
      (ns @ ns', vs @ vs')
  | PMatch (_,_,y,p1,z,p2) ->
      let (ns,vs) = collectnv p1 in 
      let (ns',vs') = collectnv p2 in 
      (ns @ ns', y :: z :: vs @ vs')

(* Check if all nonces and variables are unique *)
let uniquenv (p1, (pn, p2) :process * (process * process)) : bool = 
  let (ns1, vs1) = collectnv p1 in
  let (ns2, vs2) = collectnv p2 in
  let (ns3, vs3) = collectnv pn in
  let ns,vs = ns1 @ ns2 @ ns3, vs1 @ vs2 @ vs3 in 
  let (ns', vs') = (List.sort_uniq Pervasives.compare (List.map fst ns), List.sort_uniq Pervasives.compare vs) in 
  List.length ns = List.length ns' && List.length vs = List.length vs'

(* extract the nonce environment from the process *)
let extract_e (p1, (pn, p2) :process * (process * process)) : nonce_env = 
  let nl = fst (collectnv p1) @ fst (collectnv p2) @ fst (collectnv pn) in
  List.fold_left (fun f (n,l) -> extend f n l) env_empty nl


(* Check that replication scoping is respected *)

exception Repl_Error of string * error * error

let rec check_repl_scoping (p:process) (repl:bool) : bool = 
  match p with 
  | PZero e  -> true
  | PNew (e, n, l, q) -> 
    (match n with 
    | Nonce _ -> 
      if repl then 
        raise (Repl_Error ("Cannot bind non-star nonce under replication", e, e))
      else
        check_repl_scoping q repl
    | NonceStar n -> 
        if not repl then 
          raise (Repl_Error ("Cannot bind star-nonce without replication", e, e))
        else 
          check_repl_scoping q repl)
  | POut (e, t, q) -> 
      if not repl && not (no_star t) then 
        raise (Repl_Error ("Cannot usd star-terms without replication", e, e))
      else
        check_repl_scoping q repl
  | PIn (e, x, q) -> check_repl_scoping q repl
  | PPar (e, q1, q2) -> 
      check_repl_scoping q1 repl && check_repl_scoping q2 repl
  | PRepl (e, p') -> 
      if repl then 
        raise (Repl_Error ("Illegal nested replication", e, e))
      else
        check_repl_scoping p' true
  | PLet (e, x, d, q1, q2) -> check_repl_scoping q1 repl && check_repl_scoping q2 repl
  | PIf (e, u, v, q1, q2) -> 
      if not repl &&  not (no_star u && no_star v) then
        raise (Repl_Error ("Cannot usd star-terms without replication", e, e))
      else
        check_repl_scoping q1 repl && check_repl_scoping q2 repl

  | PMatch (e, x, y, q1, z, q2) -> check_repl_scoping q1 repl && check_repl_scoping q2 repl


(* returns the list of nonces that appear in p but are not bound in p *)
let unbound_nonces (p:process) : (nonce list) =
(* aux returns the unbound nonces in p which are not in nn *)
  let rec unbound_nonces_aux (p:process) (nn:nonce list) : (nonce list) =
    match p with
    | PZero _ -> []
    | PNew (_, n, _, p') -> unbound_nonces_aux p' (n :: nn)
    | POut (_, t, p') -> let nt = fst (collectnv_t t) in
                         let nt' = List.filter (fun n -> not (List.mem n nn)) nt in
                         nt' @ (unbound_nonces_aux p' nn)
    | PIn (_, _, p') -> unbound_nonces_aux p' nn
    | PPar (_, p', p'') -> (unbound_nonces_aux p' nn) @ (unbound_nonces_aux p'' nn)
    | PRepl (_, p') -> unbound_nonces_aux p' nn
    | PLet (_, _, _, p', p'') -> (unbound_nonces_aux p' nn) @ (unbound_nonces_aux p'' nn)
    | PIf (_, t, t', p', p'') ->
       let nt = (fst (collectnv_t t)) @ (fst (collectnv_t t')) in
       let nt' = List.filter (fun n -> not (List.mem n nn)) nt in
       nt' @ (unbound_nonces_aux p' nn) @ (unbound_nonces_aux p'' nn)
    | PMatch (_, _, _, p', _, p'') -> (unbound_nonces_aux p' nn) @ (unbound_nonces_aux p'' nn)
  in
  unbound_nonces_aux p []

(*return true iff the scoping is correct in p, ie a nonce that appears unbound is never declared new *)
let scoping_nonces (p:process) : bool =
  let unbound = unbound_nonces p in
  let bound = List.map fst (fst (collectnv p)) in
  List.for_all (fun x -> not (List.mem x bound)) unbound
  

                            
