open Syntax

let __verbose_level = ref 0
(* Utility Functions *)

(* 2string functions *)

let nonce2string (n:nonce) : string = 
  match n with | Nonce s | NonceStar s -> s

let var2string (x:var) : string = x

let rec term2string (t:term) : string  = match t with 
  | VNonce (_, n) -> nonce2string n
  | VConst (_, c) -> c
  | VVar (_, x) -> var2string x
  | VKey (_, k) -> k 
  | VPKey (_, PKey k) -> "pk("^k^")"
  | VVerKey (_, VerKey k) -> "vk("^k^")"
  | VPair (_, u, v) -> "("^ term2string u ^","^term2string v^")"
  | VInl (_, v) -> "inl("^term2string v^")"
  | VInr (_, v) -> "inr("^term2string v^")"
  | VHash (_, v) -> "h("^term2string v^")"
  | VSEnc (_, v, k) -> "enc("^term2string v^","^k^")"
  | VAEnc (_, v, PKey k) -> "aenc("^term2string v^",pk("^k^"))"
  | VSign (_, v, k) -> "sign("^term2string v^","^k^")"

let label2string (l:label) : string = match l with 
| LL -> "L"
| LH -> "H"
| LS -> "S"

let rec type2string (t:ty) : string = match t with
| TLabel l -> label2string l
| TPair (t1, t2) -> "(" ^ type2string t1 ^" * "^type2string t2 ^ ")"
| TOr (t1, t2) -> "(" ^ type2string t1 ^" v "^type2string t2 ^ ")"
| TSum (t1, t2) -> "(" ^ type2string t1 ^" + "^type2string t2 ^ ")"
| TKey (l, t) -> "key("^label2string l^", "^type2string t ^ ")"
| TRefine (u,v) -> "{x_L=" ^ term2string u ^ ", x_R="^term2string v^"}"
| TSEnc (t,k) -> "(" ^ type2string t ^ ")_" ^ k
| TAEnc (t,k) -> "{" ^ type2string t ^ "}_" ^ k

let dest2string (d:dest) : string = match d with
| DSDec (_,x, k) -> "dec("^var2string x^","^k^")"
| DADec (_,x, k) -> "adec("^var2string x^","^k^")"
| DProj1 (_,p) -> "proj_1("^var2string p^")"
| DProj2 (_,p) -> "proj_2("^var2string p^")"
| DVer (_,x, VerKey k) -> "ver("^var2string x^", vk("^k^"))"

let rec process2string (p:process) : string = match p with
| PZero e  -> "0"
| PNew (e, n, l, q) -> "new " ^ nonce2string n ^ ":" ^ label2string l ^ "." ^ process2string q
| POut (e, t, q) -> "out(" ^ term2string t ^ ")." ^ process2string q
| PIn (e, x, q) -> "in(" ^ var2string x ^ ")." ^ process2string q
| PPar (e, q1, q2) -> "(" ^ process2string q1 ^ " | " ^ process2string q2 ^ ")"
| PRepl (e, p') -> "!(" ^ process2string p' ^ ")"
| PLet (e, x, d, q1, q2) -> "let " ^ var2string x ^ "=" ^ dest2string d ^ " in " ^ process2string q1 ^ " else " ^ process2string q2
| PIf (e, u, v, q1, q2) -> "if " ^ term2string u ^ "=" ^ term2string v ^ " then " ^ process2string q1 ^ " else " ^ process2string q2
| PMatch (e, x, y, q1, z, q2) -> "match " ^ var2string x ^ " with | inl(" ^ var2string y ^ ") -> " ^ process2string q1 ^ " | inr(" ^ var2string z ^ ") -> " ^ process2string q2 

let constr2string ((u,v):constr) : string = term2string u ^" ~ "^ term2string v
let constrset2string (cs:constr_set) : string =  
  "{" ^ String.concat "\n" (List.map constr2string cs) ^ "}" 
let gamma2string (g:var_env): string =
  "[" ^
    String.concat ", "
                  (VarEnv.fold (fun x -> fun t -> fun l ->  ((var2string x ^ ":" ^ type2string t)::l)) g [])
  ^ "]"
let constrsetenv2string ((cs, g):constr_set_env) : string =  
  constrset2string cs ^ "  " ^ gamma2string g 
let constrsetset2string (css:constr_set_set) : string =  
  "{" ^ String.concat "\n\n\n" (List.map constrsetenv2string css) ^ "}"

let print_verbose (v:int) (s:string) : unit =
  if !__verbose_level >= v then print_endline s
 
  
(* equality check funcitons (necessary becasue defaultt equality is broken by
 * different error positions) *)

let rec same_term (t:term) (t':term) : bool = 
  match t,t' with 
  | VNonce (_, n), VNonce (_,n') -> n = n'
  | VConst (_, c), VConst (_,c') -> c = c'
  | VVar (_, x), VVar (_,x') -> x = x'
  | VKey (_, k), VKey (_,k') -> k = k'
  | VPKey (_, k), VPKey (_,k') -> k = k'
  | VPair (_, u, v), VPair (_,u',v') -> same_term u u' && same_term v v'
  | VInl (_, v), VInl (_,v') -> same_term v v'
  | VInr (_, v), VInr (_,v') -> same_term v v'
  | VHash (_, v), VHash (_,v') -> same_term v v'
  | VSEnc (_, v, k), VSEnc (_,v',k') -> k = k' && same_term v v'
  | VAEnc (_, v, k), VAEnc (_,v',k') -> k = k' && same_term v v'
  | VSign (_, v, k), VSign (_,v',k') -> k = k' && same_term v v'
  | _ -> false


let rec same_type (t:ty) (t':ty) : bool = 
  match t,t' with 
  | TLabel l, TLabel l'  -> l = l' 
  | TPair (t1, t2), TPair (t1',t2') -> same_type t1 t1' && same_type t2 t2'
  | TOr (t1, t2), TOr (t1',t2') -> same_type t1 t1' && same_type t2 t2'
  | TSum (t1, t2), TSum (t1',t2') -> same_type t1 t1' && same_type t2 t2'
  | TKey (l, t), TKey (l',t') -> l = l' && same_type t t'
  | TRefine (u,v), TRefine (u',v') -> same_term u u' && same_term v v'
  | TSEnc (t,k), TSEnc (t', k') -> k = k' && same_type t t'
  | TAEnc (t,k), TAEnc (t', k') -> k = k' && same_type t t'
  | _ -> false

let same_de (de:dest) (de':dest) : bool = 
match de,de' with
| DSDec (_,x, k), DSDec (_,x', k') -> x = x' && k = k'
| DADec (_,x, k), DADec (_,x', k') -> x = x' && k = k'
| DProj1 (_,p), DProj1 (_,p')-> p = p'
| DProj2 (_,p), DProj2 (_,p')-> p = p'
| DVer (_,x, VerKey k), DVer (_,x', VerKey k') -> x = x' && k = k'
| _ -> false

(* Commonly used access to subcomponents *)

let dest_var (de:dest) : var = match de with
| DSDec (_,x, k) -> x
| DADec (_,x, k) -> x
| DProj1 (_,p) -> p
| DProj2 (_,p) -> p
| DVer (_,x, VerKey k) -> x

let dest_err (de:dest) : error = match de with
| DSDec (e,x, k) -> e
| DADec (e,x, k) -> e
| DProj1 (e,p) -> e
| DProj2 (e,p) -> e
| DVer (e,x, VerKey k) -> e

let term_error (t:term) : error = match t with
| VNonce (er, _) 
| VConst (er, _) 
| VVar (er, _) 
| VKey (er, _) 
| VPKey (er, _) 
| VVerKey (er, _) 
| VPair (er,_, _) 
| VInl (er, _)
| VInr (er, _)
| VHash (er, _) 
| VSEnc (er, _, _) 
| VAEnc (er, _, _)
| VSign (er, _, _) -> er

let process_error (p:process) : error = match p with
| PZero e  -> e
| PNew (e, n, l, q) -> e
| POut (e, t, q) -> e
| PIn (e, x, q) -> e
| PPar (e, q1, q2) -> e
| PRepl (e, p') -> e
| PLet (e, x, d, q1, q2) -> e
| PIf (e, u, v, q1, q2) -> e
| PMatch (e, x, y, q1, z, q2) -> e

let is_nonce (t:term) : bool= 
  match t with 
  | VNonce _ -> true
  | _ -> false

let is_star_nonce (t:term) : bool = 
  match t with 
  | VNonce (_, (NonceStar _)) -> true
  | _ -> false

let rec no_star (t:term) : bool = 
  match t with
  | VNonce (_, n) -> not (is_star_nonce t)
  | VConst (_, c) -> true
  | VVar (_, x) -> true
  | VKey (_, k) -> true 
  | VPKey (_, PKey k) -> true
  | VVerKey (_, VerKey k) -> true
  | VPair (_, u, v) -> no_star u && no_star v
  | VInl (_, v) -> no_star v
  | VInr (_, v) -> no_star v
  | VHash (_, v) -> no_star v
  | VSEnc (_, v, k) -> no_star v
  | VAEnc (_, v, PKey k) -> no_star v
  | VSign (_, v, k) -> no_star v


(* Utility functions *)

let rec subtyping (t1:ty) (t2:ty) = 
  (* SRefl *)
  same_type t1 t2 ||
  match t1, t2 with 
  (* SHigh *)
  | _ , TLabel LH -> true
  (* SPair *)
  | TPair (t1', t1''), TPair (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SPairL *)
  | TPair (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SPairS  + SPairS'*)
  | TPair (t1', t1''), TLabel LS -> subtyping t1' (TLabel LS) ||  subtyping t1'' (TLabel LS)
  (* SSum *)
  | TSum (t1', t1''), TSum (t2', t2'') -> subtyping t1' t2' && subtyping t1'' t2''
  (* SSumL *)
  | TSum (t1', t1''), TLabel LL -> subtyping t1' (TLabel LL) && subtyping t1'' (TLabel LL)
  (* SKey *)
  | TKey (l,t), TLabel l' -> l = l'
  (* SEnc *)
  | TSEnc (t,k), TSEnc (t',k') -> k = k' &&  subtyping t t'
  (* SAEnc *)
  | TAEnc (t,k), TAEnc (t',k') -> k = k' &&  subtyping t t'
  (* Rule STrans is incomporated directly in the other rules *)
  | _ -> false
   

let env_empty (x:'a) : 'b option = None
let var_env_empty () : var_env = VarEnv.empty

let extend (env: 'a -> 'b option) (k:'a) (v:'b) (x:'a) : 'b option = if x = k then Some v else env x
let extend_nou (env: 'a -> 'b option) (k:'a) (v:'b) (x:'a) : 'b option = if env k = None then if x = k then Some v else env x else failwith "OVERWRITE"

let lookup_var (g:var_env) (x:var) : ty  option =
  try 
    Some (VarEnv.find x g)
  with
  | _ -> None 

let extend_var_env (g:var_env) (x:var) (t:ty) : var_env =
  VarEnv.add x t g

let extend_var_env_nou (g:var_env) (x:var) (t:ty) : var_env =
  match lookup_var g x with
  | Some _ -> failwith "Environment may not be updated at this point"
  | None -> extend_var_env g x t

exception UnionError 

let var_env_union (g:var_env) (g': var_env) : var_env option = 
  try
    (*we could use the ocaml 4.03 union function, but for compatibility we use merge*)
    Some (VarEnv.merge
            (fun x ot1 ot2 ->
              match ot1,ot1 with
              | Some t1, Some t2 ->
                 if same_type t1 t2 then Some t1 
                 else raise UnionError
              | None, _ -> ot2
              | _, None -> ot1) g g')
  with
  | UnionError -> None

let add_forall (c:constr) (css:constr_set_set) : constr_set_set =  
  List.map (fun (cs,g) -> (c::cs,g)) css
let union_forall (cs:constr_set) (css:constr_set_set) : constr_set_set = 
  List.map (fun (cs',g) -> (cs@cs',g)) css
let add_cart (css:constr_set_set) (css':constr_set_set) : constr_set_set = 
  List.flatten 
  (List.map (fun (cs, g) -> List.map (fun (cs', g') -> match var_env_union g g' with | Some gn -> cs@ cs', gn | None -> [], var_env_empty ()) css') css)


(* instantiate all variables with refinement types *)
let rec instantiate (g:var_env) (t:term) (left:bool) (star:bool): term = 
  let res = 
    match t with
    | VNonce (_,n) -> t
    | VConst (_,c) -> t
    | VVar (_,x) -> 
        (match lookup_var g x with 
        | Some (TRefine (t1,t2)) -> 
          if star || (no_star t1 && no_star t2) then
            if left then t1 else t2
          else
            t
        | _ -> t)
    | VKey (_,k) -> t
    | VPKey (_,PKey k) -> t
    | VVerKey (_,VerKey k) -> t
    | VPair (er,t1, t2) -> 
        let t1' = instantiate g t1 left star in 
        let t2' = instantiate g t2 left star in 
        VPair (er,t1', t2')
    | VInl (er,v) -> VInl  (er,instantiate g v left star)
    | VInr (er,v) -> VInr (er,instantiate g v left star)
    | VHash (er,v) -> VHash (er,instantiate g v left star)
    | VSEnc (er, v, k) -> VSEnc (er,(instantiate g v left star), k)
    | VAEnc (er, v, k) -> VAEnc (er,(instantiate g v left star), k)
    | VSign (er, v, k) -> VSign (er,(instantiate g v left star), k)
  in 
(*
  print_verbose 2 "Instantiate :" ;
  print_verbose 2 (term2string t ^ " to " ^ term2string res ^ "\n") ; 
*)
  res

(* Check if term contains a nonce or key with label S directly under a pair *)
let rec s_under_pair (e:nonce_env) (d:key_env) (t:term) : bool = 
  match t with 
  | VNonce (_, n) -> 
      (match e n with 
       | Some LS (*LS,_*)-> true 
       | _ -> false)
  | VKey (_, k) -> 
      (match d k with 
      | Some (TKey (LS, _)) -> true 
      | _ -> false)
  | VPair (_,t1,t2) -> s_under_pair e d t1 || s_under_pair e d t2
  | VInl (_,t1) -> s_under_pair e d t1
  | VInr (_,t2) -> s_under_pair e d t2
  | _ -> false

(* Check if a term contains a star-subterm *)


let rec contains_star_t (t:term) : bool  = match t with 
  | VNonce (_, n) -> 
      (match n with 
      | Nonce _ -> false
      | NonceStar _ -> true)
  | VConst (_, c) -> false
  | VVar (_, x) -> false
  | VKey (_, k) -> false
  | VPKey (_, PKey k) -> false
  | VVerKey (_, VerKey k) -> false
  | VPair (_, u, v) -> contains_star_t u || contains_star_t v 
  | VInl (_, v) -> contains_star_t v
  | VInr (_, v) -> contains_star_t v
  | VHash (_, v) -> contains_star_t v
  | VSEnc (_, v, k) -> contains_star_t v
  | VAEnc (_, v, PKey k) -> contains_star_t v
  | VSign (_, v, k) -> contains_star_t v

(* collect all nonces and variables of a term *)
let rec collectnv_t (t:term) : (nonce list * var list) = 
  match t with
  | VNonce (_, n) -> ([n], [])
  | VConst (_, c) -> ([], [])
  | VVar (_, x) -> ([], [x]) 
  | VKey (_, k) -> ([],[]) 
  | VPKey (_,PKey k) -> ([],[]) 
  | VVerKey (_,VerKey k) -> ([],[]) 
  | VPair (_,u, v) -> 
      let (n1,v1) = collectnv_t u in 
      let (n2,v2) = collectnv_t v in 
      (List.sort_uniq Pervasives.compare (n1 @ n2),
      List.sort_uniq Pervasives.compare (v1 @ v2))
  | VInl (_,v) -> collectnv_t v
  | VInr (_,v) -> collectnv_t v
  | VHash (_,v) -> collectnv_t v
  | VSEnc (_,v, k) -> collectnv_t v
  | VAEnc (_,v, PKey k) -> collectnv_t v
  | VSign (_,v, k) -> collectnv_t v

let ground (t:term) : bool = let (_, v) = collectnv_t t in v = []

(* collect nonces and variables from constraints *)
let collectnv_c ((u,v):constr) : nonce list * var list  = 
  let (n_u, v_u) = collectnv_t u in 
  let (n_v, v_v) = collectnv_t v in 
  List.sort_uniq Pervasives.compare (n_u @ n_v),
  List.sort_uniq Pervasives.compare (v_u @ v_v)

let well_formed_refinement (t:term) (t':term) : bool  = 
  ground t && ground t' && (is_star_nonce t && is_star_nonce t') || (no_star t && no_star t')

let rec concat_news_process (news:process) (q:process) : process =
  match news with 
  | PZero _ -> q
  | PNew (er,n,l,p) -> PNew (er,n,l, concat_news_process p q)
  | _ -> failwith "Process may only contain 'new' instructions"


(* unused *)
(* remove from g the bindings for variables which do not appear in cs *)
(*let trim_gamma (cs:constr_set) (g:var_env) : var_env =
  let vs = List.flatten (List.map (fun c -> snd (collectnv_c c)) cs) in
  VarEnv.filter (fun x -> fun t -> List.mem x vs) g*)
       
(* remove from g the binding for x if x does not appear in cs, and do nothing otherwise *)
let trim_gamma_var (cs:constr_set) (g:var_env) (x:var) : var_env =
  let vs = List.flatten (List.map (fun c -> snd (collectnv_c c)) cs) in
  if List.mem x vs then
    g
  else
    VarEnv.filter (fun y -> fun t -> x <> y) g

let remove_dup_constr_set (cs:constr_set) : constr_set =
  List.sort_uniq Pervasives.compare cs



let clean_constr_set_env_var ((cs,g):constr_set_env) (x:var) : constr_set_env =
  let cs'= remove_dup_constr_set cs in
  (cs',trim_gamma_var cs' g x)

let clean_constr_set_set_var css x = 
  let css' = List.map (fun c -> clean_constr_set_env_var c x) css in
  let css'' = List.sort_uniq Pervasives.compare css' in
  let l = (List.length css) - (List.length css'') in
  print_verbose 2 ("--> simplifying constraints: " ^ string_of_int l^" constraint(s) removed (variable: " ^ (var2string x)^")");
  css''

type side = Left | Right

(* could use a more efficient representation *)
type substitution = (var * term) list

let subs_var (s:substitution) (x:var) : term option =
  try
    Some (List.assoc x s)
  with
    Not_found -> None
               
                  
(* was previously named instantiate *)
let rec substitute (s:substitution) (u:term) : term = 
  match u with
  | VNonce (_,n) -> u
  | VConst (_,c) -> u
  | VVar (_,x) -> 
      (match subs_var s x with 
      | Some v -> v
      | None -> u)
  | VKey (_,k) -> u
  | VPKey (_,PKey k) -> u
  | VVerKey (_,VerKey k) -> u
  | VPair (er, u1, u2) -> 
      let u1' = substitute s u1 in 
      let u2' = substitute s u2 in 
      VPair (er, u1', u2')
  | VInl (er,v) -> VInl  (er, substitute s v)
  | VInr (er,v) -> VInr (er, substitute s v)
  | VHash (er,v) -> VHash (er, substitute s v)
  | VSEnc (er, v, k) -> VSEnc (er,(substitute s v), k)
  | VAEnc (er, v, k) -> VAEnc (er,(substitute s v), k)
  | VSign (er, v, k) -> VSign (er,(substitute s v), k)

